from __future__ import annotations

import sys
from pathlib import Path
from typing import Dict, List, Union

from harlequin.exception import HarlequinConfigError

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

CONFIG_FILENAMES = ("pyproject.toml", ".harlequin.toml")  # order matters!
SEARCH_DIRS = (Path.home(), Path.cwd())

Profile = Dict[str, Union[bool, int, List[str], str, Path]]
Config = Dict[str, Union[str, Dict[str, Profile]]]


def get_config_for_profile(config_path: Path | None, profile: str | None) -> Profile:
    config = load_config(config_path)
    active_profile = profile or config.get("default_profile", None)
    if active_profile is None or active_profile == "None":
        return {}
    elif active_profile not in config.get("profiles", {}):
        raise HarlequinConfigError(
            f"Could not load the profile named {active_profile} because it does not "
            "exist in any discovered config files.",
            title="Harlequin couldn't load your profile.",
        )
    else:
        return config["profiles"][active_profile]  # type: ignore


def load_config(config_path: Path | None) -> Config:
    paths = _find_config_files(config_path)
    config = _merge_config_files(paths)
    _raise_on_bad_schema(config)
    return config


def _find_config_files(config_path: Path | None) -> list[Path]:
    found_files: list[Path] = []
    if config_path is None:
        for filename in CONFIG_FILENAMES:
            for p in [p / filename for p in SEARCH_DIRS]:
                if p.exists():
                    found_files.append(p)
    elif config_path.exists():
        found_files.append(config_path)
    else:
        raise HarlequinConfigError(
            f"Config file could not be found at specified path: {config_path}",
            title="Harlequin couldn't load your config file.",
        )
    return found_files


def _merge_config_files(paths: list[Path]) -> Config:
    config: Config = {}
    for p in paths:
        try:
            with open(p, "rb") as f:
                raw_config = tomllib.load(f)
        except OSError as e:
            raise HarlequinConfigError(
                f"Error opening config file at {p}. {e}",
                title="Harlequin couldn't load your config file.",
            ) from e
        except tomllib.TOMLDecodeError as e:
            raise HarlequinConfigError(
                f"Error decoding config file at {p}. " f"Check for invalid TOML. {e}",
                title="Harlequin couldn't load your config file.",
            ) from e
        relevant_config = (
            raw_config
            if p.stem != "pyproject"
            else raw_config.get("tool", {}).get("harlequin", {})
        )
        config.update(relevant_config)
    return config


def _raise_on_bad_schema(config: Config) -> None:
    TOP_LEVEL_KEYS = ("default_profile", "profiles")
    if not config:
        return

    for k in config.keys():
        if k not in TOP_LEVEL_KEYS:
            raise HarlequinConfigError(
                f"Found unexpected key in config: {k}.\n"
                f"Allowed values are {TOP_LEVEL_KEYS}.",
                title="Harlequin couldn't load your config file.",
            )
    if config.get("profiles", None) is None:
        pass
    elif not isinstance(config["profiles"], dict):
        raise HarlequinConfigError(
            "The profiles key must define a table.",
            title="Harlequin couldn't load your config file.",
        )
    elif not all(
        [isinstance(config["profiles"][k], dict) for k in config["profiles"].keys()]
    ):
        raise HarlequinConfigError(
            "The members of the profiles table must be tables.",
            title="Harlequin couldn't load your config file.",
        )
    elif any(k == "None" for k in config["profiles"].keys()):
        raise HarlequinConfigError(
            "Config file defines a profile named 'None', which is not allowed.",
            title="Harlequin couldn't load your config file.",
        )
    else:
        for profile_name, opt_dict in config["profiles"].items():
            for option_name in opt_dict.keys():
                if "-" in option_name:
                    raise HarlequinConfigError(
                        f"Profile {profile_name} defines an option '{option_name}',"
                        "which is an invalid name for an option. Did you mean "
                        f"""'{option_name.strip("-").replace("-", "_")}'?""",
                        title="Harlequin couldn't load your config file.",
                    )

    if (default := config.get("default_profile", None)) is not None and not isinstance(
        default, str
    ):
        raise HarlequinConfigError(
            f"Config file sets default_profile to {default}, but that value "
            "must be a string.",
            title="Harlequin couldn't load your config file.",
        )
    elif (
        default is not None
        and isinstance(default, str)
        and isinstance(config["profiles"], dict)
        and default != "None"
        and config["profiles"].get(default, None) is None
    ):
        raise HarlequinConfigError(
            f"Config file sets default_profile to {default}, but does not define a "
            "profile with that name.",
            title="Harlequin couldn't load your config file.",
        )



from typing import Union

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.validation import Integer
from textual.widget import Widget
from textual.widgets import Button, Checkbox, Input


class RunQueryBar(Horizontal):
    def __init__(
        self,
        *children: Widget,
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa
        classes: Union[str, None] = None,
        disabled: bool = False,
        max_results: int = 10_000,
    ) -> None:
        self.max_results = max_results
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        yield Checkbox("Limit ", id="limit_checkbox")
        yield Input(
            str(min(500, self.max_results)),
            id="limit_input",
            validators=Integer(
                minimum=0,
                maximum=self.max_results if self.max_results > 0 else None,
                failure_description=(
                    f"Please enter a number between 0 and {self.max_results}."
                    if self.max_results > 0
                    else "Please enter a number greater than 0."
                ),
            ),
        )
        yield Button("Run Query", id="run_query")

    def set_not_responsive(self) -> None:
        self.add_class("non-responsive")

    def set_responsive(self) -> None:
        self.remove_class("non-responsive")

    def on_mount(self) -> None:
        self.checkbox = self.query_one(Checkbox)
        self.input = self.query_one(Input)
        self.button = self.query_one(Button)

    def on_input_changed(self, message: Input.Changed) -> None:
        """
        check and uncheck the box for valid/invalid input
        """
        if message.input.id == "limit_input":
            if (
                message.input.value
                and message.validation_result
                and message.validation_result.is_valid
            ):
                self.checkbox.value = True
            else:
                self.checkbox.value = False



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Sequence

import duckdb
from duckdb.typing import DuckDBPyType
from harlequin.adapter import HarlequinAdapter, HarlequinConnection, HarlequinCursor
from harlequin.catalog import Catalog, CatalogItem
from harlequin.exception import (
    HarlequinConfigError,
    HarlequinConnectionError,
    HarlequinCopyError,
    HarlequinQueryError,
)
from textual_fastdatatable.backend import AutoBackendType

from harlequin_duckdb.cli_options import DUCKDB_OPTIONS
from harlequin_duckdb.copy_formats import DUCKDB_COPY_FORMATS


class DuckDbCursor(HarlequinCursor):
    def __init__(
        self, conn: DuckDbConnection, relation: duckdb.DuckDBPyRelation
    ) -> None:
        self.conn = conn
        self.relation = relation

    def columns(self) -> list[tuple[str, str]]:
        return list(
            zip(
                self.relation.columns,
                map(self.conn._short_column_type, self.relation.dtypes),
            )
        )

    def set_limit(self, limit: int) -> HarlequinCursor:
        try:
            self.relation = self.relation.limit(limit)
        except duckdb.Error:
            pass
        return self

    def fetchall(self) -> AutoBackendType | None:
        try:
            result = self.relation.fetch_arrow_table()
        except duckdb.Error as e:
            raise HarlequinQueryError(
                msg=str(e), title="DuckDB raised an error when running your query:"
            ) from e
        return result  # type: ignore


class DuckDbConnection(HarlequinConnection):
    RELATION_TYPE_MAPPING = {
        "BASE TABLE": "t",
        "LOCAL TEMPORARY": "tmp",
        "VIEW": "v",
    }

    COLUMN_TYPE_MAPPING = {
        "SQLNULL": "\\n",
        "BOOLEAN": "t/f",
        "TINYINT": "#",
        "UTINYINT": "u#",
        "SMALLINT": "#",
        "USMALLINT": "u#",
        "INTEGER": "#",
        "UINTEGER": "u#",
        "BIGINT": "##",
        "UBIGINT": "u##",
        "HUGEINT": "###",
        "UUID": "uid",
        "FLOAT": "#.#",
        "DOUBLE": "#.#",
        "DATE": "d",
        "TIMESTAMP": "ts",
        "TIMESTAMP_MS": "ts",
        "TIMESTAMP_NS": "ts",
        "TIMESTAMP_S": "ts",
        "TIME": "t",
        "TIME_TZ": "ttz",
        "TIMESTAMP_TZ": "ttz",
        "TIMESTAMP WITH TIME ZONE": "ttz",
        "VARCHAR": "s",
        "BLOB": "0b",
        "BIT": "010",
        "INTERVAL": "|-|",
        # these types don't have python classes
        "DECIMAL": "#.#",
        "REAL": "#.#",
        "STRUCT": "{}",
        "MAP": "{m}",
    }

    UNKNOWN_TYPE = "?"

    def __init__(self, conn: duckdb.DuckDBPyConnection, init_message: str = "") -> None:
        self.conn: duckdb.DuckDBPyConnection = conn
        self.init_message = init_message

    def copy(
        self, query: str, path: Path, format_name: str, options: dict[str, Any]
    ) -> None:
        if not query:
            raise HarlequinCopyError("Cannot export result of empty query.")
        try:
            cursor = self.execute(query)
        except HarlequinQueryError as e:
            raise HarlequinCopyError(msg=e.msg, title=e.title) from e
        if cursor is None:
            raise HarlequinCopyError("Cannot export result of a DDL/DML query.")
        final_path = str(path.expanduser())
        kwargs = {k: v for k, v in options.items() if v}
        if format_name == "csv":
            if kwargs.get("quoting"):
                kwargs["quoting"] = "ALL"
            try:
                cursor.relation.write_csv(file_name=final_path, **kwargs)
            except (duckdb.Error, OSError) as e:
                raise HarlequinCopyError(
                    str(e),
                    title=(
                        "DuckDB raised an error when writing your query "
                        "to a CSV file."
                    ),
                ) from e
        elif format_name == "parquet":
            try:
                cursor.relation.write_parquet(
                    file_name=final_path, compression=kwargs.get("compression")
                )
            except (duckdb.Error, OSError) as e:
                raise HarlequinCopyError(
                    str(e),
                    title=(
                        "DuckDB raised an error when writing your query "
                        "to a Parquet file."
                    ),
                ) from e
        elif format_name == "json":
            array = f"{', ARRAY TRUE' if kwargs.get('array') else ''}"
            compression = f", COMPRESSION {kwargs.get('compression')}"
            date_format = (
                f", DATEFORMAT '{kwargs.get('''date_format''')}'"
                if kwargs.get("date_format")
                else ""
            )
            ts_format = (
                f", TIMESTAMPFORMAT '{kwargs.get('''options.timestamp_format''')}'"
                if kwargs.get("options.timestamp_format")
                else ""
            )
            try:
                self.execute(
                    f"copy ({query}) to '{final_path}' "
                    "(FORMAT JSON"
                    f"{array}{compression}{date_format}{ts_format}"
                    ")"
                )
            except (HarlequinQueryError, OSError) as e:
                raise HarlequinCopyError(
                    str(e),
                    title=(
                        "DuckDB raised an error when writing your query "
                        "to a JSON file."
                    ),
                ) from e

    def execute(self, query: str) -> DuckDbCursor | None:
        try:
            rel = self.conn.sql(query)
        except duckdb.Error as e:
            raise HarlequinQueryError(
                msg=str(e),
                title="DuckDB raised an error when compiling or running your query:",
            ) from e

        if rel is not None:
            return DuckDbCursor(conn=self, relation=rel)
        else:
            return None

    def get_catalog(self) -> Catalog:
        catalog_items: list[CatalogItem] = []
        databases = self._get_databases()
        for (database,) in databases:
            database_identifier = f'"{database}"'
            schemas = self._get_schemas(database)
            schema_items: list[CatalogItem] = []
            for (schema,) in schemas:
                schema_identifier = f'{database_identifier}."{schema}"'
                tables = self._get_tables(database, schema)
                table_items: list[CatalogItem] = []
                for table, kind in tables:
                    table_identifier = f'{schema_identifier}."{table}"'
                    columns = self._get_columns(database, schema, table)
                    column_items = [
                        CatalogItem(
                            qualified_identifier=f'{table_identifier}."{col[0]}"',
                            query_name=f'"{col[0]}"',
                            label=col[0],
                            type_label=self._short_column_type(col[1]),
                        )
                        for col in columns
                    ]
                    table_items.append(
                        CatalogItem(
                            qualified_identifier=table_identifier,
                            query_name=table_identifier,
                            label=table,
                            type_label=self._short_relation_type(kind),
                            children=column_items,
                        )
                    )
                schema_items.append(
                    CatalogItem(
                        qualified_identifier=schema_identifier,
                        query_name=schema_identifier,
                        label=schema,
                        type_label="sch",
                        children=table_items,
                    )
                )
            catalog_items.append(
                CatalogItem(
                    qualified_identifier=database_identifier,
                    query_name=database_identifier,
                    label=database,
                    type_label="db",
                    children=schema_items,
                )
            )
        return Catalog(items=catalog_items)

    def validate_sql(self, text: str) -> str:
        cur = self.conn.cursor()
        escaped = text.replace("'", "''")
        try:
            (parsed,) = cur.sql(  # type: ignore
                f"select json_serialize_sql('{escaped}')"
            ).fetchone()
        except HarlequinQueryError:
            return ""
        result = json.loads(parsed)
        # DDL statements return an error of type "not implemented"
        if result.get("error", True) and result.get("error_type", "") == "parser":
            return ""
        else:
            return text

    def _get_databases(self) -> list[tuple[str]]:
        cur = self.conn.cursor()
        return cur.execute("pragma show_databases").fetchall()

    def _get_schemas(self, database: str) -> list[tuple[str]]:
        cur = self.conn.cursor()
        schemas = cur.execute(
            "select schema_name "
            "from information_schema.schemata "
            "where "
            "    catalog_name = ? "
            "    and schema_name not in ('pg_catalog', 'information_schema') "
            "order by 1",
            [database],
        ).fetchall()
        return schemas

    def _get_tables(self, database: str, schema: str) -> list[tuple[str, str]]:
        cur = self.conn.cursor()
        tables = cur.execute(
            "select table_name, table_type "
            "from information_schema.tables "
            "where "
            "    table_catalog = ? "
            "    and table_schema = ? "
            "order by 1",
            [database, schema],
        ).fetchall()
        return tables

    def _get_columns(
        self, database: str, schema: str, table: str
    ) -> list[tuple[str, str]]:
        cur = self.conn.cursor()
        columns = cur.execute(
            "select column_name, data_type "
            "from information_schema.columns "
            "where "
            "    table_catalog = ? "
            "    and table_schema = ? "
            "    and table_name = ? "
            "order by 1",
            [database, schema, table],
        ).fetchall()
        return columns

    @classmethod
    def _short_relation_type(cls, native_type: str) -> str:
        return cls.RELATION_TYPE_MAPPING.get(native_type, cls.UNKNOWN_TYPE)

    @classmethod
    def _short_column_type(cls, native_type: DuckDBPyType | str) -> str:
        """
        In duckdb v0.8.0, relation.dtypes started returning a DuckDBPyType,
        instead of a string. However, this type isn't an ENUM, and there
        aren't classes for all types, so it's hard
        to check class members. So we just convert to a string and split
        complex types on their first paren to match our dictionary.
        """
        base_type = str(native_type).split("(")[0].split("[")[0]
        short_base_type = cls.COLUMN_TYPE_MAPPING.get(base_type, cls.UNKNOWN_TYPE)
        if str(native_type).endswith("[]"):
            return f"[{short_base_type}]"
        else:
            return short_base_type


class DuckDbAdapter(HarlequinAdapter):
    ADAPTER_OPTIONS = DUCKDB_OPTIONS
    COPY_FORMATS = DUCKDB_COPY_FORMATS

    def __init__(
        self,
        conn_str: Sequence[str],
        init_path: Path | str | None = None,
        no_init: bool | str = False,
        read_only: bool | str = False,
        allow_unsigned_extensions: bool | str = False,
        extension: list[str] | None = None,
        force_install_extensions: bool = False,
        custom_extension_repo: str | None = None,
        md_token: str | None = None,
        md_saas: bool = False,
        **_: Any,
    ) -> None:
        try:
            self.conn_str = conn_str if conn_str else (":memory:",)
            self.init_path = (
                Path(init_path).resolve()
                if init_path is not None
                else Path.home() / ".duckdbrc"
            )
            self.no_init = bool(no_init)
            self.read_only = bool(read_only)
            self.allow_unsigned_extensions = bool(allow_unsigned_extensions)
            self.extensions = extension if extension is not None else []
            self.force_install_extensions = force_install_extensions
            self.custom_extension_repo = custom_extension_repo
            self.md_token = md_token
            self.md_saas = md_saas
        except (ValueError, TypeError) as e:
            raise HarlequinConfigError(
                msg=f"DuckDB adapter received bad config value: {e}",
                title="Harlequin could not initialize the selected adapter.",
            ) from e

    def connect(self) -> DuckDbConnection:
        primary_db, *other_dbs = self.conn_str
        token = f"?token={self.md_token}" if self.md_token else ""
        saas = "?saas_mode=true" if self.md_saas else ""
        config = {
            "allow_unsigned_extensions": str(self.allow_unsigned_extensions).lower()
        }

        try:
            connection = duckdb.connect(
                database=f"{primary_db}{token}{saas}",
                read_only=self.read_only,
                config=config,
            )
            for db in other_dbs:
                connection.execute(
                    f"attach '{db}'{' (READ_ONLY)' if self.read_only else ''}"
                )
        except (duckdb.CatalogException, duckdb.IOException) as e:
            raise HarlequinConnectionError(
                str(e), title="DuckDB couldn't connect to your database."
            ) from e

        if self.custom_extension_repo:
            connection.execute(
                f"SET custom_extension_repository='{self.custom_extension_repo}';"
            )

        if self.extensions:
            try:
                for extension in self.extensions:
                    # todo: support installing from a URL instead.
                    connection.install_extension(
                        extension=extension, force_install=self.force_install_extensions
                    )
                    connection.load_extension(extension=extension)
            except (duckdb.HTTPException, duckdb.IOException) as e:
                raise HarlequinConnectionError(
                    str(e), title="DuckDB couldn't install or load your extension."
                ) from e

        msg = ""
        if self.init_path is not None and not self.no_init:
            init_script = self._read_init_script(self.init_path)
            try:
                count = 0
                for command in self._split_script(init_script):
                    rewritten_command = self._rewrite_init_command(command)
                    for cmd in rewritten_command.split(";"):
                        if cmd.strip():
                            connection.execute(cmd)
                            count += 1
            except duckdb.Error as e:
                msg = f"Attempted to execute script at {init_script[0]}\n{e}"
                raise HarlequinConnectionError(
                    msg, title="DuckDB could not execute your initialization script."
                ) from e
            else:
                if count > 0:
                    msg = (
                        f"Executed {count} {'command' if count == 1 else 'commands'} "
                        f"from {self.init_path}"
                    )

        return DuckDbConnection(conn=connection, init_message=msg)

    @staticmethod
    def _read_init_script(init_path: Path) -> str:
        try:
            with open(init_path.expanduser(), "r") as f:
                init_script = f.read()
        except OSError:
            init_script = ""
        return init_script

    @staticmethod
    def _split_script(script: str) -> list[str]:
        """
        DuckDB init scripts can contain SQL queries or dot commands. The SQL
        queries may contain newlines, but the dot commands are newline-terminated.
        This takes a raw script and returns a list of executable commands
        """
        lines = script.splitlines()
        commands: list[str] = []
        i = 0
        for j, line in enumerate(lines):
            if line.startswith("."):
                commands.append("\n".join(lines[i:j]))
                commands.append(line)
                i = j + 1
        commands.append("\n".join(lines[i:]))
        return [command.strip() for command in commands if command]

    @staticmethod
    def _rewrite_dot_open(command: str) -> str:
        """
        Rewrites .open command into its SQL equivalent.
        """
        args = command.split()[1:]
        if not args:
            return "attach ':memory:'; use memory;"
        else:
            # --readonly is only supported option
            if len(args) == 2 and args[0] == "--readonly":
                option = " (READ_ONLY)"
                db_path = Path(args[1])
            else:
                option = ""
                db_path = Path(args[0])

            return f"attach '{db_path}'{option} as {db_path.stem}; use {db_path.stem};"

    @classmethod
    def _rewrite_init_command(cls, command: str) -> str:
        """
        DuckDB init scripts can contain dot commands, which can only be executed
        by the CLI, not the python API. Here, we rewrite some common ones into
        SQL, and rewrite the others to be no-ops.
        """
        if not command.startswith("."):
            return command
        elif command.startswith(".open"):
            return cls._rewrite_dot_open(command)
        else:
            return ""



from typing import Dict, Union

from pygments.style import Style as PygmentsStyle
from pygments.styles import get_style_by_name
from pygments.token import (
    Comment,
    Error,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
    Token,
)
from pygments.util import ClassNotFound
from questionary import Style as QuestionaryStyle
from textual.design import ColorSystem

from harlequin.exception import HarlequinThemeError

GREEN = "#45FFCA"
YELLOW = "#FEFFAC"
PINK = "#FFB6D9"
PURPLE = "#D67BFF"
GRAY = "#777777"
DARK_GRAY = "#333333"
BLACK = "#0C0C0C"
WHITE = "#DDDDDD"


class HarlequinPygmentsStyle(PygmentsStyle):
    styles = {
        Token: WHITE,
        Comment: f"{GRAY} italic",
        Keyword: f"{YELLOW} bold",
        Keyword.Type: f"{YELLOW} nobold",
        Keyword.Constant: f"{PINK} bold",
        Name: WHITE,
        Name.Builtin: YELLOW,
        Name.Constant: YELLOW,
        Name.Quoted: f"{WHITE} bold",
        Name.Variable: f"{WHITE} bold",
        String: PINK,
        String.Symbol: f"{WHITE} bold",
        String.Name: f"{WHITE} bold",
        Operator: GREEN,
        Punctuation: PURPLE,
        Number: f"{PINK} bold",
        Literal: PINK,
        Error: PINK,
    }
    background_color = BLACK
    highlight_color = DARK_GRAY


HARLEQUIN_QUESTIONARY_STYLE = QuestionaryStyle(
    [
        ("qmark", f"fg:{GREEN} bold"),
        ("question", "bold"),
        ("answer", f"fg:{YELLOW} bold"),
        ("pointer", f"fg:{YELLOW} bold"),
        ("highlighted", f"fg:{YELLOW} bold"),
        ("selected", f"fg:{YELLOW} noreverse bold"),
        ("separator", f"fg:{PURPLE}"),
        ("instruction", "fg:#858585 italic"),
        ("text", ""),
        ("disabled", "fg:#858585 italic"),
    ]
)


def extract_color(s: str) -> str:
    for part in s.split(" "):
        if part.startswith("#"):
            return part
    else:
        return ""


class HarlequinColorSystem(ColorSystem):
    def __init__(
        self,
        primary: str,
        secondary: Union[str, None] = None,
        warning: Union[str, None] = None,
        error: Union[str, None] = None,
        success: Union[str, None] = None,
        accent: Union[str, None] = None,
        background: Union[str, None] = None,
        surface: Union[str, None] = None,
        panel: Union[str, None] = None,
        boost: Union[str, None] = None,
        dark: bool = False,
        luminosity_spread: float = 0.15,
        text_alpha: float = 0.95,
        text: Union[str, None] = None,
    ):
        super().__init__(
            primary,
            secondary,
            warning,
            error,
            success,
            accent,
            background,
            surface,
            panel,
            boost,
            dark,
            luminosity_spread,
            text_alpha,
        )
        self.text = text

    def generate(self) -> Dict[str, str]:
        colors = super().generate()
        if self.text:
            colors["text"] = self.text
        return colors


class HarlequinColors:
    MAPPING = {
        "text": [Token, Token.Name, Token.Name.Variable],
        "primary": [Token.Keyword, Token],
        "secondary": [
            Token.Name.Function,
            Token.Name.Class,
            Token.Literal.String,
            Token,
        ],
        "gray": [Token.Comment, Token],
        "error": [Token.Error, Token.Name.Exception, Token],
    }
    FALLBACK = {
        "text": "#000000",
        "primary": "#000000",
        "secondary": "#000000",
        "gray": "#888888",
        "error": "#FF0000",
    }

    def __init__(
        self,
        background: str,
        highlight: str,
        text: str,
        primary: str,
        secondary: str,
        gray: str,
        error: str,
    ) -> None:
        self.background = background
        self.highlight = highlight
        self.text = text
        self.primary = primary
        self.secondary = secondary
        self.gray = gray
        self.error = error

    @classmethod
    def from_theme(cls, theme: str) -> "HarlequinColors":
        try:
            style = get_style_by_name(theme)
        except ClassNotFound as e:
            raise HarlequinThemeError(
                (
                    f"No theme found with the name {theme}.\n"
                    "Theme must be the name of a Pygments Style. "
                    "You can browse the supported styles here:\n"
                    "https://pygments.org/styles/"
                ),
                title="Harlequin couldn't load your theme.",
            ) from e

        background = style.background_color
        highlight = style.highlight_color
        style_colors = {
            prop_name: [
                extract_color(style.styles.get(token_type, ""))
                for token_type in token_type_list
                if extract_color(style.styles.get(token_type, ""))
            ]
            for prop_name, token_type_list in cls.MAPPING.items()
        }
        best_colors = {k: v[0] for k, v in style_colors.items() if v}
        best_colors.update(
            {k: cls.FALLBACK[k] for k, v in style_colors.items() if not v}
        )
        return cls(background, highlight, **best_colors)

    @property
    def color_system(self) -> ColorSystem:
        return ColorSystem(
            primary=self.primary,
            secondary=self.secondary,
            error=self.error,
            background=self.background,
            surface=self.background,
            boost=self.highlight,
            panel=self.gray,
        )

    @property
    def design_system(self) -> Dict[str, ColorSystem]:
        return {
            "dark": self.color_system,
            "light": self.color_system,
        }



from __future__ import annotations

import sqlite3
from itertools import zip_longest
from pathlib import Path
from typing import Any, Literal, Sequence
from urllib.parse import unquote, urlparse

from harlequin.adapter import HarlequinAdapter, HarlequinConnection, HarlequinCursor
from harlequin.catalog import Catalog, CatalogItem
from harlequin.exception import (
    HarlequinConfigError,
    HarlequinConnectionError,
    HarlequinQueryError,
)
from harlequin.options import HarlequinAdapterOption, HarlequinCopyFormat
from textual_fastdatatable.backend import AutoBackendType

from harlequin_sqlite.cli_options import SQLITE_OPTIONS


class HarlequinSqliteCursor(HarlequinCursor):
    def __init__(self, conn: HarlequinSqliteConnection, cur: sqlite3.Cursor) -> None:
        self.conn = conn
        self.cur = cur
        self._limit: int | None = None
        _first_row = cur.fetchone()
        self.has_records = _first_row is not None
        self._first_row: tuple[Any, ...] = _first_row or tuple(
            [None] * len(cur.description)
        )

    def columns(self) -> list[tuple[str, str]]:
        col_names = [col[0] for col in self.cur.description]
        col_types = [
            self.conn._short_column_type_from_python_object(value)
            for value in self._first_row
        ]
        return list(zip_longest(col_names, col_types, fillvalue="?"))

    def set_limit(self, limit: int) -> "HarlequinSqliteCursor":
        self._limit = limit
        return self

    def fetchall(self) -> AutoBackendType | None:
        if self.has_records:
            remaining_rows = (
                self.cur.fetchall()
                if self._limit is None
                else self.cur.fetchmany(self._limit - 1)
            )
            return [self._first_row, *remaining_rows]
        else:
            return None


class HarlequinSqliteConnection(HarlequinConnection):
    def __init__(self, conn: sqlite3.Connection, init_message: str = "") -> None:
        self.conn = conn
        self.init_message = init_message

    def execute(self, query: str) -> HarlequinSqliteCursor | None:
        try:
            cur = self.conn.execute(query)
        except sqlite3.Error as e:
            raise HarlequinQueryError(
                msg=str(e),
                title="SQLite raised an error when compiling or running your query:",
            ) from e

        if cur.description is not None:
            return HarlequinSqliteCursor(conn=self, cur=cur)
        else:
            return None

    def get_catalog(self) -> Catalog:
        catalog_items: list[CatalogItem] = []
        databases = self._get_databases()
        for db_name in databases:
            relations = self._get_relations(db_name=db_name)
            db_identifier = f'"{db_name}"'
            rel_items: list[CatalogItem] = []
            for rel_name, rel_type in relations:
                rel_identifier = f'{db_identifier}."{rel_name}"'
                cols = self._get_columns(db_name=db_name, rel_name=rel_name)
                col_items = [
                    CatalogItem(
                        qualified_identifier=f'{rel_identifier}."{col_name}"',
                        query_name=f'"{col_name}"',
                        label=col_name,
                        type_label=self._short_column_type(col_type),
                    )
                    for _, col_name, col_type, *_ in cols
                ]
                rel_items.append(
                    CatalogItem(
                        qualified_identifier=rel_identifier,
                        query_name=rel_identifier,
                        label=rel_name,
                        type_label=self._short_relation_type(rel_type),
                        children=col_items,
                    )
                )
            catalog_items.append(
                CatalogItem(
                    qualified_identifier=db_identifier,
                    query_name=db_identifier,
                    label=db_name,
                    type_label="db",
                    children=rel_items,
                )
            )
        return Catalog(items=catalog_items)

    def _get_databases(self) -> list[str]:
        objects: list[tuple[str, str, str]] = self.conn.execute(
            "pragma database_list"
        ).fetchall()
        return [db_name for _, db_name, _ in objects]

    def _get_relations(self, db_name: str) -> list[tuple[str, str]]:
        objects = self.conn.execute(
            f'select type, name from "{db_name}".sqlite_schema'
        ).fetchall()
        relations = [(name, typ) for typ, name in objects if typ in ("table", "view")]
        return relations

    def _get_columns(self, db_name: str, rel_name: str) -> list[tuple[str, str, str]]:
        return self.conn.execute(
            f"pragma {db_name}.table_info('{rel_name}')"
        ).fetchall()

    @staticmethod
    def _short_column_type(raw_type: str) -> str:
        # first get the storage affinity for a type decl
        if raw_type is None or raw_type == "":
            affinity = ""

        typ = raw_type.lower()
        if "int" in typ:
            affinity = "INTEGER"
        elif any([x in typ for x in ["char", "clob", "text"]]):
            affinity = "TEXT"
        elif "blob" in typ:
            affinity = "BLOB"
        elif any([x in typ for x in ["real", "floa", "doub"]]):
            affinity = "REAL"
        else:
            affinity = "NUMERIC"

        # then lookup the abbreviation for the affinity
        mapping = {
            "": "",
            "TEXT": "s",
            "NUMERIC": "#.#",
            "INTEGER": "##",
            "REAL": "#.#",
            "BLOB": "b",
        }
        return mapping.get(affinity, "?")

    @staticmethod
    def _short_column_type_from_python_object(obj: object) -> str:
        mapping: dict[type | None, str] = {
            None: "",
            str: "s",
            float: "#.#",
            int: "##",
            object: "b",
        }
        return mapping.get(type(obj), "?")

    @staticmethod
    def _short_relation_type(raw_type: str) -> str:
        mapping = {"table": "t", "view": "v"}
        return mapping.get(raw_type, "?")

    def copy(
        self, query: str, path: Path, format_name: str, options: dict[str, Any]
    ) -> None:
        raise NotImplementedError

    def validate_sql(self, text: str) -> str:
        raise NotImplementedError


class HarlequinSqliteAdapter(HarlequinAdapter):
    ADAPTER_OPTIONS: list[HarlequinAdapterOption] | None = SQLITE_OPTIONS
    COPY_FORMATS: list[HarlequinCopyFormat] | None = None

    def __init__(
        self,
        conn_str: Sequence[str],
        read_only: bool = False,
        connection_mode: Literal["ro", "rw", "rwc", "memory"] | None = None,
        timeout: str | float = 5.0,
        detect_types: str | int = 0,
        isolation_level: Literal["DEFERRED", "EXCLUSIVE", "IMMEDIATE"]
        | None = "DEFERRED",
        cached_statements: str | int = 128,
        **_: Any,
    ) -> None:
        try:
            self.conn_str = conn_str if conn_str else (":memory:",)
            self.read_only = bool(read_only)
            self.connection_mode = connection_mode
            self.timeout = float(timeout)
            self.detect_types = int(detect_types)
            self.isolation_level = isolation_level
            self.cached_statements = int(cached_statements)
        except (ValueError, TypeError) as e:
            raise HarlequinConfigError(
                msg=f"SQLite adapter received bad config value: {e}",
                title="Harlequin could not initialize the selected adapter.",
            ) from e

    def connect(self) -> HarlequinSqliteConnection:
        if (
            self.read_only
            and self.connection_mode is not None
            and self.connection_mode != "ro"
        ):
            raise HarlequinConnectionError(
                "Cannot specify readonly flag and a connection mode."
            )
        elif self.read_only:
            mode_str = "?mode=ro"
        elif self.connection_mode is not None:
            mode_str = f"?mode={self.connection_mode}"
        else:
            mode_str = ""

        # build db strings
        db_uris: list[str] = []
        db_names: list[str] = []
        for s in self.conn_str:
            if s == ":memory:":
                db_uris.append(s)
                db_names.append("memory")
            elif s.startswith("file:"):
                db_uris.append(s)
                db_names.append(Path(unquote(urlparse(s).path)).stem)
            else:
                try:
                    p = Path(s).resolve()
                    db = f"{p.as_uri()}{mode_str}"
                    db_uris.append(db)
                    db_names.append(p.stem)
                except ValueError as e:
                    raise HarlequinConnectionError(
                        f"Cannot build URI from connection string {s}.",
                        title="SQLite couldn't connect to your database.",
                    ) from e
        primary_db, *other_dbs = db_uris
        try:
            conn = sqlite3.connect(
                database=primary_db,
                timeout=self.timeout,
                detect_types=self.detect_types,
                isolation_level=self.isolation_level,
                cached_statements=self.cached_statements,
                check_same_thread=False,
                uri=True,
            )
        except Exception as e:
            raise HarlequinConnectionError(
                msg=str(e), title="SQLite couldn't connect to your database."
            ) from e

        for uri, name in zip(other_dbs, db_names[1:]):
            try:
                conn.execute(f"attach database '{uri}' as {name}")
            except sqlite3.Error as e:
                raise HarlequinConnectionError(
                    msg=str(e), title="SQLite couldn't connect to your database {name}."
                ) from e
        return HarlequinSqliteConnection(conn=conn)



from typing import Union

import pyperclip
from textual import events
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Static


class ClickableStatic(Static):
    def on_click(self, message: events.Click) -> None:
        message.stop()
        try:
            pyperclip.copy(str(self.renderable))
        except pyperclip.PyperclipException:
            pass
        else:
            self.app.notify("Error copied to clipboard.")


class ErrorModal(ModalScreen):
    def __init__(
        self,
        title: str,
        header: str,
        error: BaseException,
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa: A002
        classes: Union[str, None] = None,
    ) -> None:
        super().__init__(name, id, classes)
        self.title = title
        self.header = header
        self.error = error

    def compose(self) -> ComposeResult:
        with Vertical(id="error_outer"):
            yield Static(self.header, id="error_header")
            with Vertical(id="error_inner"):
                with VerticalScroll():
                    yield ClickableStatic(str(self.error), id="error_info")
            yield Static(
                "Press any key to continue. Click error to copy.", id="error_footer"
            )

    def on_mount(self) -> None:
        container = self.query_one("#error_outer")
        container.border_title = self.title

    def on_key(self) -> None:
        self.app.pop_screen()

    def on_click(self, message: events.Click) -> None:
        self.app.pop_screen()



from typing import Generic, List, Set, Tuple, Union

from textual.binding import Binding
from textual.events import Click
from textual.message import Message
from textual.widgets import Tree
from textual.widgets._tree import EventTreeDataType, TreeNode

from harlequin.catalog import Catalog, CatalogItem


class DataCatalog(Tree[CatalogItem]):
    BINDINGS = [
        Binding(
            "ctrl+enter",
            "submit",
            "Insert Name",
            key_display="CTRL+ENTER / CTRL+J",
            show=True,
        ),
        Binding("ctrl+j", "submit", "Insert Name", show=False),
    ]

    BORDER_TITLE = "Data Catalog"

    class NodeSubmitted(Generic[EventTreeDataType], Message):
        def __init__(self, node: TreeNode[EventTreeDataType]) -> None:
            self.node: TreeNode[EventTreeDataType] = node
            super().__init__()

    def __init__(
        self,
        type_color: str = "#888888",
        data: Union[CatalogItem, None] = None,
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa: A002
        classes: Union[str, None] = None,
        disabled: bool = False,
    ) -> None:
        self.type_color = type_color
        super().__init__(
            "Root", data, name=name, id=id, classes=classes, disabled=disabled
        )
        self.double_click: Union[int, None] = None

    def update_tree(self, catalog: Catalog) -> None:
        tree_state = self._get_node_states(self.root)
        expanded_nodes: Set[str] = set(tree_state[0])
        # todo: tree's select_node() not working
        # unless the tree is modified, the selection will stay
        # in the same place
        # selected_node = tree_state[1]
        self.clear()
        if catalog.items:
            self._build_subtree(catalog.items, self.root, expanded_nodes)

    def on_mount(self) -> None:
        self.show_root = False
        self.guide_depth = 3
        self.root.expand()

    async def on_click(self, event: Click) -> None:
        """
        For whatver reason, it doesn't seem possible to override the super class's
        _on_click event. Instead, here we just handle the parts relevant
        to double clicking, and to prevent nodes from collapsing after double-clicking,
        we collapse them, since the _on_click event will go after this and toggle
        their state.
        """
        meta = event.style.meta
        click_line: Union[int, None] = meta.get("line", None)
        if (
            self.double_click is not None
            and click_line is not None
            and self.double_click == click_line
        ):
            node = self.get_node_at_line(click_line)
            if node is not None:
                self.post_message(self.NodeSubmitted(node=node))
                node.collapse()
        else:
            self.double_click = click_line
            self.set_timer(
                delay=0.5, callback=self._clear_double_click, name="double_click_timer"
            )

    def action_submit(self) -> None:
        if isinstance(self.cursor_line, int) and self.cursor_line > -1:
            node = self.get_node_at_line(self.cursor_line)
            if node is not None:
                self.post_message(self.NodeSubmitted(node=node))

    def _build_item_label(self, label: str, type_label: str) -> str:
        return f"{label} [{self.type_color}]{type_label}[/]" if type_label else label

    def _build_subtree(
        self,
        items: List[CatalogItem],
        parent: TreeNode[CatalogItem],
        expanded_nodes: Set[str],
    ) -> None:
        for item in items:
            if item.children:
                new_node = parent.add(
                    label=self._build_item_label(item.label, item.type_label),
                    data=item,
                    expand=item.qualified_identifier in expanded_nodes,
                )
                self._build_subtree(item.children, new_node, expanded_nodes)
            else:
                parent.add_leaf(
                    label=self._build_item_label(item.label, item.type_label), data=item
                )

    def _clear_double_click(self) -> None:
        self.double_click = None

    @classmethod
    def _get_node_states(
        cls, node: TreeNode[CatalogItem]
    ) -> Tuple[List[str], Union[str, None]]:
        expanded_nodes = []
        selected_node = None
        if node.is_expanded and node.data is not None:
            expanded_nodes.append(node.data.qualified_identifier)
        if node._selected and node.data is not None:
            selected_node = node.data.qualified_identifier
        for child in node.children:
            expanded_children, selected_child = cls._get_node_states(child)
            expanded_nodes.extend(expanded_children)
            selected_node = selected_child or selected_node
        return expanded_nodes, selected_node



from textual.widgets import Input as Input
from textual.widgets import Label
from textual.widgets import Select as Select
from textual.widgets import Switch as Switch
from textual_textarea import PathInput as PathInput


class NoFocusLabel(Label, can_focus=False):
    pass



from __future__ import annotations

from pathlib import Path
from typing import Any

import questionary
import tomlkit
from pygments.styles import get_all_styles
from rich import print as rich_print
from rich.panel import Panel
from rich.syntax import Syntax
from tomlkit.exceptions import TOMLKitError
from tomlkit.toml_document import TOMLDocument
from tomlkit.toml_file import TOMLFile

from harlequin.adapter import HarlequinAdapter
from harlequin.colors import HARLEQUIN_QUESTIONARY_STYLE, YELLOW
from harlequin.exception import HarlequinWizardError, pretty_print_error
from harlequin.options import ListOption
from harlequin.plugins import load_plugins


def wizard() -> None:
    try:
        _wizard()
    except KeyboardInterrupt:
        print("Cancelled config updates. No changes were made to any files.")
        return
    except HarlequinWizardError as e:
        pretty_print_error(e)
        return


def _wizard() -> None:
    path, is_pyproject = _prompt_for_path()
    config, file = _read_toml(path)

    # extract existing profiles from config file.
    if is_pyproject:
        full_config = config
        config = config.get("tool", {}).get("harlequin", {})
    if "profiles" not in config:
        config["profiles"] = {}
    profiles = config.get("profiles", {})

    profile_name = _prompt_for_profile_name(profiles)
    selected_profile = profiles.get(profile_name, {})

    adapters = load_plugins()
    adapter = questionary.select(
        message="Which adapter should this profile use?",
        choices=sorted(adapters.keys()),
        default=selected_profile.get("adapter", "duckdb"),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    conn_str = questionary.text(
        message="What connection string(s) should this profile use?",
        instruction="Separate items by a space.",
        default=" ".join(selected_profile.get("conn_str", [])),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    theme = questionary.select(
        message="What theme should this profile use?",
        choices=list(get_all_styles()),
        default=selected_profile.get("theme", "monokai"),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    limit = int(
        questionary.text(
            message="How many rows should the data table show?",
            validate=_validate_int,
            default=str(selected_profile.get("limit", 100000)),
            style=HARLEQUIN_QUESTIONARY_STYLE,
        ).unsafe_ask()
    )

    adapter_cls = adapters[adapter]
    adapter_option_choices = (
        [
            questionary.Choice(
                title=opt.name, checked=_sluggify_name(opt.name) in selected_profile
            )
            for opt in adapter_cls.ADAPTER_OPTIONS
        ]
        if adapter_cls.ADAPTER_OPTIONS is not None
        else []
    )
    which = questionary.checkbox(
        message="Which of the following adapter options would you like to set?",
        choices=adapter_option_choices,
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    adapter_options = {}
    if conn_str:
        adapter_options["conn_str"] = conn_str.split(" ")
    _prompt_to_set_adapter_options(
        adapter_options=adapter_options,
        adapter_cls=adapter_cls,
        which=which,
        selected_profile=selected_profile,
    )

    default_profile = _prompt_to_set_default_profile(profile_name, config, profiles)

    new_profile = {
        "adapter": adapter,
        "theme": theme,
        "limit": limit,
        **adapter_options,
    }

    _confirm_profile_generation(default_profile, profile_name, new_profile)

    config["profiles"][profile_name] = new_profile  # type: ignore

    if is_pyproject:
        if "tool" not in full_config:
            full_config["tool"] = {}
        full_config["tool"]["harlequin"] = config  # type: ignore
        config = full_config

    file.write(config)


def _prompt_for_path() -> tuple[Path, bool]:
    raw_path: str = questionary.path(
        "What config file do you want to create or update?",
        default=".harlequin.toml",
        validate=lambda p: True
        if p.endswith(".toml")
        else "Must have a .toml extension",
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()
    path = Path(raw_path)
    is_pyproject = path.stem == "pyproject"
    if path.suffix != ".toml":
        raise HarlequinWizardError(
            msg="Must create a file with a .toml extension.",
            title="Harlequin could not create your configuration.",
        )
    return path, is_pyproject


def _read_toml(path: Path) -> tuple[TOMLDocument, TOMLFile]:
    file = TOMLFile(path)
    try:
        config = file.read()
    except OSError:
        config = TOMLDocument()
    except TOMLKitError as e:
        raise HarlequinWizardError(
            f"Attempted to load the config file at {path}, but encountered an "
            f"error:\n\n{e}",
            title="Harlequin could not load the config file.",
        ) from e
    return config, file


def _prompt_for_profile_name(profiles: TOMLDocument) -> str:
    NEW_PROFILE_SENTINEL = "[Create a New Profile]"
    profile_name = NEW_PROFILE_SENTINEL
    if profiles:
        profile_name = questionary.select(
            message="Which profile would you like to update?",
            choices=[NEW_PROFILE_SENTINEL, *profiles.keys()],
            style=HARLEQUIN_QUESTIONARY_STYLE,
        ).unsafe_ask()
    if profile_name == NEW_PROFILE_SENTINEL:
        profile_name = questionary.text(
            message="What would you like to name your profile?",
            style=HARLEQUIN_QUESTIONARY_STYLE,
            validate=lambda x: True if x and x != "None" else "Cannot be empty or None",
        ).unsafe_ask()
    return profile_name


def _prompt_to_set_adapter_options(
    adapter_options: dict[str, Any],
    adapter_cls: type[HarlequinAdapter],
    which: list[str],
    selected_profile: TOMLDocument,
) -> None:
    """
    Mutates passed adapter_options dict.
    """
    if which and adapter_cls.ADAPTER_OPTIONS is not None:
        for option in adapter_cls.ADAPTER_OPTIONS:
            if option.name not in which:
                continue
            value = option.to_questionary(
                selected_profile.get(_sluggify_name(option.name), None)
            ).unsafe_ask()
            if isinstance(option, ListOption):
                value = value.split(" ")
            adapter_options.update({_sluggify_name(option.name): value})


def _prompt_to_set_default_profile(
    profile_name: str, config: TOMLDocument, profiles: TOMLDocument
) -> str | None:
    possible_names = set([profile_name, *profiles.keys()])
    NO_DEFAULT_SENTINEL = "[No default]"
    default_profile: str = questionary.select(
        message="Would you like to set a default profile?",
        choices=[
            NO_DEFAULT_SENTINEL,
            *possible_names,
        ],
        default=config.get("default_profile", None),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    if default_profile == NO_DEFAULT_SENTINEL:
        _ = config.pop("default_profile", None)
        return None
    else:
        config["default_profile"] = default_profile
        return default_profile


def _confirm_profile_generation(
    default_profile: str | None, profile_name: str, new_profile: dict[str, Any]
) -> None:
    new_config: dict[str, Any] = (
        {} if default_profile is None else {"default_profile": default_profile}
    )
    new_config.update({"profiles": {profile_name: new_profile}})
    new_config_toml = tomlkit.dumps(new_config).rstrip()

    rich_print("[italic] We generated the following profile:[/]")
    rich_print(
        Panel.fit(
            Syntax(code=new_config_toml, lexer="toml", theme="harlequin"),
            border_style=YELLOW,
        )
    )

    all_good = questionary.confirm(
        "Save this profile?",
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).ask()

    if not all_good:
        raise KeyboardInterrupt()


def _validate_int(raw: str) -> bool:
    try:
        int(raw)
    except ValueError:
        return False
    else:
        return True


def _sluggify_name(raw: str) -> str:
    return raw.strip("-").replace("-", "_")



from dataclasses import dataclass, field
from typing import List

from textual.message import Message


@dataclass
class CatalogItem:
    """
    A representation of a database object.

    Args:
        qualified_identifier (str): The fully-scoped and (optionally) quoted database
            identifier for this object. Should be globally unique for the Connection.
            e.g., a postgres table: "mydb"."myschema"."mytable"
        query_name (str): The text to be inserted into the query editor for this
            object.
        label (str): The unqualified name for this object, to appear in the Data
            Catalog widget.
        type_label (str): A short (1-3 chars) label denoting the type of this object
            or the data it contains. e.g., ("t") for "table" or "##" for
            "bigint column".
        children (list[CatalogItem]): Other nodes nested under this one (e.g.,
            a list of columns if this is a table.)
    """

    qualified_identifier: str
    query_name: str
    label: str
    type_label: str
    children: List["CatalogItem"] = field(default_factory=list)


@dataclass
class Catalog:
    """
    A representation of the queryable objects in a database, to be displayed in the
    DataCatalog widget.

    Args:
        items (list[CatalogItem]): A list of CatalogItem nodes, which each represent
            a database object, like a database, schema, table, or column.
    """

    items: List[CatalogItem]


class NewCatalog(Message):
    def __init__(self, catalog: Catalog) -> None:
        self.catalog = catalog
        super().__init__()



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any, Callable, Sequence

import rich_click as click

from harlequin import Harlequin
from harlequin.adapter import HarlequinAdapter
from harlequin.colors import GREEN, PINK, PURPLE, YELLOW
from harlequin.config import get_config_for_profile
from harlequin.config_wizard import wizard
from harlequin.exception import HarlequinConfigError, pretty_print_error
from harlequin.plugins import load_plugins

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points, version
else:
    from importlib.metadata import entry_points, version

# configure defaults
DEFAULT_ADAPTER = "duckdb"
DEFAULT_LIMIT = 100_000
DEFAULT_THEME = "monokai"

# configure the rich click interface (mostly --help options)
DOCS_URL = "https://harlequin.sh/docs/getting-started"

# general
click.rich_click.USE_RICH_MARKUP = True
click.rich_click.COLOR_SYSTEM = "truecolor"

click.rich_click.STYLE_OPTIONS_TABLE_LEADING = 1
click.rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE"
click.rich_click.STYLE_OPTIONS_PANEL_BORDER = YELLOW
click.rich_click.STYLE_USAGE = f"bold {YELLOW}"
click.rich_click.STYLE_USAGE_COMMAND = "regular"
click.rich_click.STYLE_HELPTEXT = "regular"
click.rich_click.STYLE_OPTION = PINK
click.rich_click.STYLE_ARGUMENT = PINK
click.rich_click.STYLE_COMMAND = PINK
click.rich_click.STYLE_SWITCH = GREEN

# metavars
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.STYLE_METAVAR_APPEND = PURPLE
click.rich_click.STYLE_METAVAR_SEPARATOR = PURPLE

# errors
click.rich_click.STYLE_ERRORS_SUGGESTION = "italic"
click.rich_click.ERRORS_SUGGESTION = "Try 'harlequin --help' to view available options."
click.rich_click.ERRORS_EPILOGUE = (
    f"To learn more, visit [link={DOCS_URL}]{DOCS_URL}[/link]"
)

# define main option group (adapter options added to own groups below)
click.rich_click.OPTION_GROUPS = {
    "harlequin": [
        {
            "name": "Harlequin Options",
            "options": [
                "--profile",
                "--config-path",
                "--adapter",
                "--theme",
                "--limit",
                "--config",
                "--version",
                "--help",
            ],
        },
    ]
}


def _version_option() -> str:
    """
    Build the string printed by harlequin --version
    """
    harlequin_version = version("harlequin")
    adapter_eps = entry_points(group="harlequin.adapter")
    adapter_versions: dict[str, str] = {}
    for ep in adapter_eps:
        adapter_versions.update({ep.name: ep.dist.version})

    adapter_output = "\n".join(
        [f"  - {name}, version {version}" for name, version in adapter_versions.items()]
    )

    output = (
        f"harlequin, version {harlequin_version}\n\n"
        "Installed Adapters:\n"
        f"{adapter_output}"
    )

    return output


def _config_wizard_callback(ctx: click.Context, param: Any, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return
    wizard()
    ctx.exit(0)


def build_cli() -> click.Command:
    """
    Loads installed adapters and constructs a click Command that includes options
    defined by all adapters.

    Returns: click.Command
    """
    adapters = load_plugins()

    @click.command()
    @click.version_option(package_name="harlequin", message=_version_option())
    @click.argument(
        "conn_str",
        nargs=-1,
    )
    @click.option(
        "-P",
        "--profile",
        help=(
            "Select a profile from an available config file to load its values. "
            "Other options passed here will take precedence over those loaded "
            "from the profile. Use the special profile named None to use Harlequin's "
            "defaults, instead of the default profile specified in the config "
            "file."
        ),
    )
    @click.option(
        "--config-path",
        help=(
            "By default, Harlequin finds files named .harlequin.toml in the "
            "current directory and the home directory (~) and merges them. "
            "Use this option to specify the full path to a config file at "
            "a different location."
        ),
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
            path_type=Path,
        ),
    )
    @click.option(
        "-t",
        "--theme",
        default=DEFAULT_THEME,
        show_default=True,
        help=(
            "Set the theme (colors) of the Harlequin IDE. "
            "Must be the name of a Pygments style; see "
            "https://pygments.org/styles/"
        ),
    )
    @click.option(
        "--limit",
        "-l",
        default=DEFAULT_LIMIT,
        type=click.IntRange(min=0),
        help=(
            "Set the maximum number of rows that can be loaded into Harlequin's "
            "Results Viewer. Set to 0 for no limit. Default is 100,000"
        ),
    )
    @click.option(
        "-a",
        "--adapter",
        default=DEFAULT_ADAPTER,
        show_default=True,
        type=click.Choice(list(adapters.keys()), case_sensitive=False),
        help=(
            "The name of an installed database adapter plug-in "
            "to use to connect to the database at CONN_STR."
        ),
    )
    @click.option(
        "--config",
        help=(
            "Run the configuration wizard to create or update a Harlequin "
            "config file."
        ),
        is_flag=True,
        callback=_config_wizard_callback,
        expose_value=True,
        is_eager=True,
    )
    @click.pass_context
    def inner_cli(
        ctx: click.Context,
        profile: str | None,
        config_path: Path | None,
        **kwargs: Any,
    ) -> None:
        """
        This command starts the Harlequin IDE.

        [bold #FFB6D9]CONN_STR[/] [#D67BFF](TEXT MULTIPLE)[/][dim]: One or more
        connection strings (or paths to local db files) for databases to open with
        Harlequin.[/]
        """
        # load config from any config files
        try:
            config = get_config_for_profile(config_path=config_path, profile=profile)
        except HarlequinConfigError as e:
            pretty_print_error(e)
            ctx.exit(2)

        # prune the kwargs to only those that don't have their default arguments
        params = list(kwargs.keys())
        for k in params:
            if (
                ctx.get_parameter_source(k)
                == click.core.ParameterSource.DEFAULT  # type: ignore
            ):
                kwargs.pop(k)
            # conn_str is an arg, not an option, so get_paramter_source is always CLI
            elif k == "conn_str" and kwargs[k] == tuple():
                kwargs.pop(k)

        # merge the config and the cli options
        config.update(kwargs)

        # load and instantiate the adapter
        adapter = config.pop("adapter", DEFAULT_ADAPTER)
        conn_str: Sequence[str] = config.pop("conn_str", tuple())  # type: ignore
        if isinstance(conn_str, str):
            conn_str = (conn_str,)
        adapter_cls: type[HarlequinAdapter] = adapters[adapter]  # type: ignore
        try:
            adapter_instance = adapter_cls(conn_str=conn_str, **config)
        except HarlequinConfigError as e:
            pretty_print_error(e)
            ctx.exit(2)

        tui = Harlequin(
            adapter=adapter_instance,
            max_results=config.get("limit", DEFAULT_LIMIT),  # type: ignore
            theme=config.get("theme", DEFAULT_THEME),  # type: ignore
        )
        tui.run()

    # iterate through installed adapters and decorate the inner_cli command
    # with the additional options declared by the adapter.
    # we load the options into a dict keyed by their name to de-dupe options
    # that may be passed by multiple adapters.
    options: dict[str, Callable[[click.Command], click.Command]] = {}
    for adapter_name, adapter_cls in adapters.items():
        option_name_list: list[str] = []
        if adapter_cls.ADAPTER_OPTIONS is not None:
            for option in adapter_cls.ADAPTER_OPTIONS:
                options.update({option.name: option.to_click()})
                option_name_list.append(f"--{option.name}")
        click.rich_click.OPTION_GROUPS["harlequin"].append(
            {"name": f"{adapter_name} Adapter Options", "options": option_name_list}
        )

    fn = inner_cli
    for click_opt in options.values():
        fn = click_opt(fn)  # type: ignore

    return fn


def harlequin() -> None:
    """
    The main entrypoint for the Harlequin IDE. Builds and executes the click Command.
    """
    cli = build_cli()
    cli()



from harlequin_duckdb.adapter import DuckDbAdapter as DuckDbAdapter
from harlequin_duckdb.cli_options import DUCKDB_OPTIONS as DUCKDB_OPTIONS
from harlequin_duckdb.copy_formats import DUCKDB_COPY_FORMATS as DUCKDB_COPY_FORMATS



class HarlequinExit(Exception):
    pass


class HarlequinError(Exception):
    def __init__(self, msg: str, title: str = "") -> None:
        super().__init__(msg)
        self.msg = msg
        self.title = title


class HarlequinConnectionError(HarlequinError):
    pass


class HarlequinCopyError(HarlequinError):
    pass


class HarlequinQueryError(HarlequinError):
    pass


class HarlequinThemeError(HarlequinError):
    pass


class HarlequinConfigError(HarlequinError):
    pass


class HarlequinWizardError(HarlequinError):
    pass


def pretty_print_error(error: HarlequinError) -> None:
    from rich import print
    from rich.panel import Panel

    print(
        Panel.fit(
            str(error),
            title=error.title if error.title else ("Harlequin encountered an error."),
            title_align="left",
            border_style="red",
        )
    )



from harlequin.components.code_editor import CodeEditor, EditorCollection
from harlequin.components.data_catalog import DataCatalog
from harlequin.components.error_modal import ErrorModal
from harlequin.components.export_screen import ExportScreen, export_callback
from harlequin.components.help_screen import HelpScreen
from harlequin.components.results_viewer import ResultsViewer
from harlequin.components.run_query_bar import RunQueryBar

__all__ = [
    "CodeEditor",
    "EditorCollection",
    "ErrorModal",
    "ExportScreen",
    "export_callback",
    "HelpScreen",
    "ResultsViewer",
    "RunQueryBar",
    "DataCatalog",
]



from harlequin.options import FlagOption, HarlequinCopyFormat, SelectOption, TextOption

csv = HarlequinCopyFormat(
    name="csv",
    label="CSV",
    extensions=(".csv", ".tsv"),
    options=[
        FlagOption(
            name="header",
            description="Switch on to include column name headers.",
            label="Header",
        ),
        TextOption(
            name="sep",
            description="The separator (or delimeter) between cols in each row.",
            label="Separator",
            default=",",
        ),
        SelectOption(
            name="compression",
            description=(
                "The compression type for the file. By default this will be detected "
                "automatically from the file extension (e.g. file.csv.gz will use "
                "gzip, file.csv will use no compression)."
            ),
            default="auto",
            choices=[
                ("Auto", "auto"),
                ("gzip", "gzip"),
                ("zstd", "zstd"),
                ("No compression", "none"),
            ],
        ),
        FlagOption(
            name="quoting",
            description="Switch on to always quote all strings.",
            label="Force Quote",
        ),
        TextOption(
            name="date_format",
            description="Specifies the date format to use when writing dates.",
            label="Date Format",
            default="",
            placeholder="%Y-%m-%d",
        ),
        TextOption(
            name="timestamp_format",
            description="Specifies the date format to use when writing timestamps.",
            label="Timestamp Format",
            default="",
            placeholder="%c",
        ),
        TextOption(
            name="quotechar",
            description="The quoting character to be used when a data value is quoted.",
            label="Quote Char",
            default='"',
        ),
        TextOption(
            name="escapechar",
            description=(
                "The character that should appear before a character that matches the "
                "quote value."
            ),
            label="Escape Char",
            default='"',
        ),
        TextOption(
            name="na_rep",
            description="The string that is written to represent a NULL value.",
            label="Null String",
            default="",
        ),
        TextOption(
            name="encoding",
            description="Only UTF8 is currently supported by DuckDB.",
            label="Encoding",
            default="UTF8",
        ),
    ],
)

parquet = HarlequinCopyFormat(
    name="parquet",
    label="Parquet",
    extensions=(".parquet", ".pq"),
    options=[
        SelectOption(
            name="compression",
            description=(
                "The compression format to use (uncompressed, snappy, gzip or zstd). "
                "Default snappy."
            ),
            choices=[
                ("Snappy", "snappy"),
                ("gzip", "gzip"),
                ("zstd", "zstd"),
                ("Uncompressed", "uncompressed"),
            ],
            default="snappy",
        )
    ],
)

json = HarlequinCopyFormat(
    name="json",
    label="JSON",
    extensions=(".json", ".js", ".ndjson"),
    options=[
        FlagOption(
            name="array",
            description=(
                "Whether to write a JSON array. If true, a JSON array of records is "
                "written, if false, newline-delimited JSON is written."
            ),
        ),
        SelectOption(
            name="compression",
            description=(
                "The compression type for the file. By default this will be detected "
                "automatically from the file extension (e.g. file.json.gz will use "
                "gzip, file.json will use no compression)."
            ),
            choices=[
                ("Auto", "auto"),
                ("gzip", "gzip"),
                ("zstd", "zstd"),
                ("Uncompressed", "uncompressed"),
            ],
            default="auto",
        ),
        TextOption(
            name="date_format",
            description="Specifies the date format to use when writing dates.",
            label="Date Format",
            default="",
            placeholder="%Y-%m-%d",
        ),
        TextOption(
            name="timestamp_format",
            description="Specifies the date format to use when writing timestamps.",
            label="Timestamp Format",
            default="",
            placeholder="%c",
        ),
    ],
)

DUCKDB_COPY_FORMATS = [csv, parquet, json]



from __future__ import annotations

import sys
from typing import Dict

from harlequin.adapter import HarlequinAdapter

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


def load_plugins() -> Dict[str, type[HarlequinAdapter]]:
    adapter_eps = entry_points(group="harlequin.adapter")
    adapters: Dict[str, type[HarlequinAdapter]] = {}

    for ep in adapter_eps:
        try:
            adapters.update({ep.name: ep.load()})
        except ImportError as e:
            print(
                f"Harlequin could not load the installed plug-in named {ep.name}."
                f"\n\n{e}"
            )

    return adapters



from typing import List, Tuple, Union

from rich.markup import escape
from textual.app import ComposeResult
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.widget import Widget
from textual.widgets import (
    ContentSwitcher,
    LoadingIndicator,
    TabbedContent,
    TabPane,
    Tabs,
)
from textual_fastdatatable import DataTable
from textual_fastdatatable.backend import AutoBackendType


class ResultsTable(DataTable):
    DEFAULT_CSS = """
        ResultsTable {
            height: 100%;
            width: 100%;
        }
    """


class ResultsViewer(ContentSwitcher, can_focus=True):
    BINDINGS = [
        Binding("j", "switch_tab(-1)", "Previous Tab", show=False),
        Binding("k", "switch_tab(1)", "Next Tab", show=False),
    ]

    TABBED_ID = "tabs"
    LOADING_ID = "loading"

    def __init__(
        self,
        *children: Widget,
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa A002
        classes: Union[str, None] = None,
        disabled: bool = False,
        initial: Union[str, None] = None,
        max_results: int = 10_000,
        type_color: str = "#888888",
    ) -> None:
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            initial=initial,
        )
        self.max_results = max_results
        self.type_color = type_color

    def compose(self) -> ComposeResult:
        yield TabbedContent(id=self.TABBED_ID)
        yield LoadingIndicator(id=self.LOADING_ID)

    def clear_all_tables(self) -> None:
        self.tab_switcher.clear_panes()
        self.add_class("hide-tabs")

    def get_visible_table(self) -> Union[ResultsTable, None]:
        content = self.tab_switcher.query_one(ContentSwitcher)
        active_tab_id = self.tab_switcher.active
        if active_tab_id:
            try:
                tab_pane = content.query_one(f"#{active_tab_id}", TabPane)
                return tab_pane.query_one(ResultsTable)
            except NoMatches:
                return None
        else:
            tables = content.query(ResultsTable)
            try:
                return tables.first(ResultsTable)
            except NoMatches:
                return None

    async def push_table(
        self,
        table_id: str,
        column_labels: List[Tuple[str, str]],
        data: AutoBackendType,
    ) -> None:
        formatted_labels = [
            self._format_column_label(col_name, col_type)
            for col_name, col_type in column_labels
        ]
        self.app.log(f"LABELS: {formatted_labels}")
        table = ResultsTable(
            id=table_id,
            column_labels=formatted_labels,  # type: ignore
            data=data,
            max_rows=self.max_results,
        )
        n = self.tab_switcher.tab_count + 1
        if n > 1:
            self.remove_class("hide-tabs")
        pane = TabPane(f"Result {n}", table)
        await self.tab_switcher.add_pane(pane)  # type: ignore
        self.tab_switcher.active = f"tab-{n}"

    async def set_not_responsive(self) -> None:
        self.border_title = "Loading Data"
        self.add_class("non-responsive")

    def set_responsive(
        self,
        did_run: bool = True,
    ) -> None:
        if not did_run:
            self.border_title = "Query Results"
        else:
            table = self.get_visible_table()
            if table is not None:
                rows = table.source_row_count
                if rows > 0:
                    self.border_title = (
                        f"Query Results {self._human_row_count(table.source_row_count)}"
                    )
                else:
                    self.border_title = "Query Returned No Records"
            else:
                self.border_title = "Query Results"
        self.remove_class("non-responsive")

    def show_loading(self) -> None:
        self.current = self.LOADING_ID
        self.border_title = "Running Query"
        self.add_class("non-responsive")

    def show_table(self) -> None:
        self.current = self.TABBED_ID

    def on_mount(self) -> None:
        self.border_title = "Query Results"
        self.current = self.TABBED_ID
        self.tab_switcher = self.query_one(TabbedContent)
        self.loading_spinner = self.query_one(LoadingIndicator)
        self.query_one(Tabs).can_focus = False

    def on_focus(self) -> None:
        self._focus_on_visible_table()

    def on_tabbed_content_tab_activated(
        self, message: TabbedContent.TabActivated
    ) -> None:
        message.stop()
        # Don't update the border if we're still loading the table.
        if self.border_title and str(self.border_title).startswith("Loading"):
            return
        maybe_table = self.get_visible_table()
        if maybe_table is not None:
            self.border_title = (
                f"Query Results {self._human_row_count(maybe_table.source_row_count)}"
            )
            maybe_table.focus()

    def action_switch_tab(self, offset: int) -> None:
        if not self.tab_switcher.active:
            return
        tab_number = int(self.tab_switcher.active.split("-")[1])
        unsafe_tab_number = tab_number + offset
        if unsafe_tab_number < 1:
            new_tab_number = self.tab_switcher.tab_count
        elif unsafe_tab_number > self.tab_switcher.tab_count:
            new_tab_number = 1
        else:
            new_tab_number = unsafe_tab_number
        self.tab_switcher.active = f"tab-{new_tab_number}"
        self._focus_on_visible_table()

    def _focus_on_visible_table(self) -> None:
        maybe_table = self.get_visible_table()
        if maybe_table is not None:
            maybe_table.focus()

    def _human_row_count(self, total_rows: int) -> str:
        if self.max_results > 0 and total_rows > self.max_results:
            return f"(Showing {self.max_results:,} of {total_rows:,} Records)"
        else:
            return f"({total_rows:,} Records)"

    def _format_column_label(self, col_name: str, col_type: str) -> str:
        return f"{escape(col_name)} [{self.type_color}]{escape(col_type)}[/]"



from __future__ import annotations

import asyncio
import time
from functools import partial
from typing import Dict, List, Optional, Type, Union

from textual import work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.css.stylesheet import Stylesheet
from textual.dom import DOMNode
from textual.driver import Driver
from textual.message import Message
from textual.reactive import reactive
from textual.screen import Screen, ScreenResultCallbackType, ScreenResultType
from textual.types import CSSPathType
from textual.widget import AwaitMount, Widget
from textual.widgets import Button, Checkbox, Footer, Input
from textual.worker import Worker, WorkerState

from harlequin.adapter import HarlequinAdapter, HarlequinCursor
from harlequin.cache import BufferState, Cache, write_cache
from harlequin.catalog import CatalogItem, NewCatalog
from harlequin.colors import HarlequinColors
from harlequin.components import (
    CodeEditor,
    DataCatalog,
    EditorCollection,
    ErrorModal,
    ExportScreen,
    HelpScreen,
    ResultsViewer,
    RunQueryBar,
    export_callback,
)
from harlequin.exception import (
    HarlequinConfigError,
    HarlequinConnectionError,
    HarlequinQueryError,
    HarlequinThemeError,
    pretty_print_error,
)


class QueriesExecuted(Message):
    def __init__(self, query_text: str, cursors: Dict[str, HarlequinCursor]) -> None:
        self.query_text = query_text
        self.cursors = cursors
        super().__init__()


class ResultsFetched(Message):
    def __init__(
        self,
        cursors: Dict[str, HarlequinCursor],
        errors: List[BaseException],
        elapsed: Union[float, None],
    ) -> None:
        self.cursors = cursors
        self.errors = errors
        self.elapsed = elapsed
        super().__init__()


class Harlequin(App, inherit_bindings=False):
    """
    The SQL IDE for your Terminal.
    """

    CSS_PATH = "global.tcss"

    BINDINGS = [
        Binding("ctrl+q", "quit", "Quit"),
        Binding("f1", "show_help_screen", "Help"),
        Binding("f2", "focus_query_editor", "Focus Query Editor", show=False),
        Binding("f5", "focus_results_viewer", "Focus Results Viewer", show=False),
        Binding("f6", "focus_data_catalog", "Focus Data Catalog", show=False),
        Binding("ctrl+b", "toggle_sidebar", "Toggle Sidebar", show=False),
        Binding("f9", "toggle_sidebar", "Toggle Sidebar", show=False),
        Binding("f10", "toggle_full_screen", "Toggle Full Screen Mode", show=False),
        Binding("ctrl+e", "export", "Export Data", show=False),
    ]

    query_text: reactive[str] = reactive(str)
    selection_text: reactive[str] = reactive(str)
    cursors: reactive[Dict[str, HarlequinCursor]] = reactive(dict)
    full_screen: reactive[bool] = reactive(False)
    sidebar_hidden: reactive[bool] = reactive(False)

    def __init__(
        self,
        adapter: HarlequinAdapter,
        theme: str = "monokai",
        max_results: int | str = 100_000,
        driver_class: Union[Type[Driver], None] = None,
        css_path: Union[CSSPathType, None] = None,
        watch_css: bool = False,
    ):
        super().__init__(driver_class, css_path, watch_css)
        self.adapter = adapter
        self.theme = theme
        try:
            self.max_results = int(max_results)
        except ValueError:
            pretty_print_error(
                HarlequinConfigError(
                    f"limit={max_results!r} was set by config file but is not "
                    "a valid integer."
                )
            )
            self.exit(return_code=2)
        else:
            self.limit = min(500, self.max_results) if self.max_results > 0 else 500
        self.query_timer: Union[float, None] = None
        try:
            self.connection = self.adapter.connect()
        except HarlequinConnectionError as e:
            pretty_print_error(e)
            self.exit(return_code=2)
        else:
            if self.connection.init_message:
                self.notify(self.connection.init_message)

        try:
            self.app_colors = HarlequinColors.from_theme(theme)
        except HarlequinThemeError as e:
            pretty_print_error(e)
            self.exit(return_code=2)
        else:
            self.design = self.app_colors.design_system
            self.stylesheet = Stylesheet(variables=self.get_css_variables())

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        with Horizontal():
            yield DataCatalog(type_color=self.app_colors.gray)
            with Vertical(id="main_panel"):
                yield EditorCollection(language="sql", theme=self.theme)
                yield RunQueryBar(max_results=self.max_results)
                yield ResultsViewer(
                    max_results=self.max_results, type_color=self.app_colors.gray
                )
        yield Footer()

    def push_screen(  # type: ignore
        self,
        screen: Union[Screen[ScreenResultType], str],
        callback: Union[ScreenResultCallbackType[ScreenResultType], None] = None,
        wait_for_dismiss: bool = False,
    ) -> Union[AwaitMount, asyncio.Future[ScreenResultType]]:
        if self.editor._has_focus_within:
            self.editor.text_input.blink_timer.pause()
        return super().push_screen(  # type: ignore
            screen,
            callback=callback,
            wait_for_dismiss=wait_for_dismiss,
        )

    def pop_screen(self) -> Screen[object]:
        new_screen = super().pop_screen()
        if len(self.screen_stack) == 1 and self.editor._has_focus_within:
            self.editor.text_input.blink_timer.resume()
        return new_screen

    async def on_mount(self) -> None:
        self.data_catalog = self.query_one(DataCatalog)
        self.editor_collection = self.query_one(EditorCollection)
        self.editor = self.editor_collection.current_editor
        self.results_viewer = self.query_one(ResultsViewer)
        self.run_query_bar = self.query_one(RunQueryBar)
        self.footer = self.query_one(Footer)

        self.editor.focus()
        self.run_query_bar.checkbox.value = False

        self.update_schema_data()

    def on_button_pressed(self, message: Button.Pressed) -> None:
        message.stop()
        if message.control.id == "run_query":
            self.query_text = self._get_query_text()

    def on_code_editor_submitted(self, message: CodeEditor.Submitted) -> None:
        message.stop()
        self.query_text = self._get_query_text()

    def on_data_catalog_node_submitted(
        self, message: DataCatalog.NodeSubmitted[CatalogItem]
    ) -> None:
        message.stop()
        if message.node.data:
            self.editor.insert_text_at_selection(text=message.node.data.query_name)
            self.editor.focus()

    def on_editor_collection_editor_switched(
        self, message: EditorCollection.EditorSwitched
    ) -> None:
        if message.active_editor is not None:
            self.editor = message.active_editor
        else:
            self.editor = self.editor_collection.current_editor

    def on_text_area_cursor_moved(self) -> None:
        self.selection_text = self._validate_selection()

    def on_checkbox_changed(self, message: Checkbox.Changed) -> None:
        """
        invalidate the last query so we re-run the query with the limit
        """
        if message.checkbox.id == "limit_checkbox":
            message.stop()
            self.query_text = ""

    def on_input_changed(self, message: Input.Changed) -> None:
        """
        invalidate the last query so we re-run the query with the limit
        """
        if message.input.id == "limit_input":
            message.stop()
            if (
                message.input.value
                and message.validation_result
                and message.validation_result.is_valid
            ):
                self.query_text = ""
                self.limit = int(message.input.value)
                message.input.tooltip = None
            elif message.validation_result:
                failures = "\n".join(message.validation_result.failure_descriptions)
                message.input.tooltip = (
                    f"[{self.app_colors.error}]Validation Error:[/]\n{failures}"
                )

    def on_input_submitted(self, message: Input.Submitted) -> None:
        if message.input.id == "limit_input":
            message.stop()
            if (
                message.input.value
                and message.validation_result
                and message.validation_result.is_valid
            ):
                self.query_text = self._get_query_text()

    def on_worker_state_changed(self, message: Worker.StateChanged) -> None:
        if message.state == WorkerState.CANCELLED:
            self._on_worker_cancelled(message)
        elif message.state == WorkerState.ERROR:
            self._on_worker_error(message)

    def _on_worker_cancelled(self, message: Worker.StateChanged) -> None:
        pass

    def _on_worker_error(self, message: Worker.StateChanged) -> None:
        if (
            message.worker.name == "_post_updated_catalog"
            and message.worker.error is not None
        ):
            self._push_error_modal(
                title="Catalog Error",
                header="Could not update data catalog",
                error=message.worker.error,
            )
        elif (
            message.worker.name == "_build_cursors" and message.worker.error is not None
        ):
            self.run_query_bar.set_responsive()
            self.results_viewer.set_responsive()
            self.results_viewer.show_table()
            header = getattr(message.worker.error, "title", "Query Error")
            self._push_error_modal(
                title="Query Error",
                header=header,
                error=message.worker.error,
            )

    def on_new_catalog(self, message: NewCatalog) -> None:
        self.data_catalog.update_tree(message.catalog)

    def on_queries_executed(self, message: QueriesExecuted) -> None:
        self.results_viewer.clear_all_tables()
        cursors = message.cursors
        number_of_queries = len(self._split_query_text(message.query_text))
        elapsed = (
            time.monotonic() - self.query_timer if self.query_timer is not None else 0.0
        )
        if cursors:  # select query
            self.cursors = cursors
            if len(cursors) < number_of_queries:
                # mixed select and DDL statements
                n = number_of_queries - len(cursors)
                self.notify(
                    f"{n} DDL/DML {'query' if n == 1 else 'queries'} "
                    f"executed successfully in {elapsed:.2f} seconds."
                )
                self.update_schema_data()
        elif bool(message.query_text.strip()):  # DDL/DML queries only
            self.run_query_bar.set_responsive()
            self.results_viewer.set_responsive()
            self.results_viewer.show_table()
            self.notify(
                f"{number_of_queries} DDL/DML "
                f"{'query' if number_of_queries == 1 else 'queries'} "
                f"executed successfully in {elapsed:.2f} seconds."
            )
            self.query_timer = None
            self.update_schema_data()
        else:  # blank query
            self.run_query_bar.set_responsive()
            self.results_viewer.set_responsive(did_run=False)
            self.results_viewer.show_table()

    def on_results_fetched(self, message: ResultsFetched) -> None:
        if message.errors:
            header = getattr(
                message.errors[0],
                "title",
                "The database raised an error when running your query:",
            )
            self._push_error_modal(
                title="Query Error",
                header=header,
                error=message.errors[0],
            )
        elif message.elapsed is not None:
            self.notify(
                f"{len(message.cursors)} "
                f"{'query' if len(message.cursors) == 1 else 'queries'} "
                f"executed successfully in {message.elapsed:.2f} seconds."
            )
        self.run_query_bar.set_responsive()
        self.results_viewer.show_table()
        if len(message.errors) == len(message.cursors):
            self.results_viewer.set_responsive(did_run=False)
        else:
            self.results_viewer.set_responsive(did_run=True)
            self.results_viewer.focus()

    def watch_full_screen(self, full_screen: bool) -> None:
        full_screen_widgets = [self.editor_collection, self.results_viewer]
        other_widgets = [self.run_query_bar, self.footer]
        all_widgets = [*full_screen_widgets, *other_widgets]
        if full_screen:
            target: Optional[DOMNode] = self.focused
            while target not in full_screen_widgets:
                if (
                    target is None
                    or target in other_widgets
                    or not isinstance(target, Widget)
                ):
                    return
                else:
                    target = target.parent
            for w in all_widgets:
                w.disabled = w != target
            if target == self.editor_collection:
                self.run_query_bar.disabled = False
            self.data_catalog.disabled = True
        else:
            for w in all_widgets:
                w.disabled = False
            self.data_catalog.disabled = self.sidebar_hidden

    async def watch_query_text(self, query_text: str) -> None:
        if query_text:
            self.query_timer = time.monotonic()
            self.full_screen = False
            self.run_query_bar.set_not_responsive()
            self.results_viewer.show_loading()
            self._build_cursors(query_text)

    async def watch_cursors(self, cursors: Dict[str, HarlequinCursor]) -> None:
        """
        Only runs for select statements, except when first mounted.
        """
        # invalidate results so watch_data runs even if the results are the same
        self.results_viewer.clear_all_tables()
        self._set_result_viewer_data(cursors)

    def watch_selection_text(self, selection_text: str) -> None:
        if selection_text:
            self.run_query_bar.button.label = "Run Selection"
        else:
            self.run_query_bar.button.label = "Run Query"

    def watch_sidebar_hidden(self, sidebar_hidden: bool) -> None:
        if sidebar_hidden:
            if self.data_catalog.has_focus:
                self.editor.focus()
        self.data_catalog.disabled = sidebar_hidden

    def action_export(self) -> None:
        if not self.adapter.implements_copy:
            self._push_error_modal(
                title="Not Implemented by Adapter",
                header="Harlequin could not load the Export menu.",
                error=NotImplementedError(
                    "The current adapter has not implemented the export menu. "
                    "Please write a COPY statement using the query editor."
                ),
            )
            return
        show_export_error = partial(
            self._push_error_modal,
            "Export Data Error",
            "Could not export data.",
        )
        queries = self._split_query_text(self._get_query_text())
        if not queries or not queries[0]:
            show_export_error(
                error=ValueError(
                    "You must type a query into the editor first, or move your "
                    "cursor into an existing query."
                )
            )
            return
        notify = partial(self.notify, "Data exported successfully.")
        callback = partial(
            export_callback,
            query=queries[0],
            connection=self.connection,
            success_callback=notify,
            error_callback=show_export_error,
        )
        self.app.push_screen(
            ExportScreen(adapter=self.adapter, id="export_screen"), callback
        )

    def action_focus_data_catalog(self) -> None:
        if self.sidebar_hidden or self.data_catalog.disabled:
            self.action_toggle_sidebar()
        self.data_catalog.focus()

    def action_focus_query_editor(self) -> None:
        self.editor.focus()

    def action_focus_results_viewer(self) -> None:
        self.results_viewer.focus()

    async def action_quit(self) -> None:
        buffers = []
        for i, editor in enumerate(self.editor_collection.all_editors):
            if editor == self.editor_collection.current_editor:
                focus_index = i
            buffers.append(
                BufferState(editor.cursor, editor.selection_anchor, editor.text)
            )
        write_cache(Cache(focus_index=focus_index, buffers=buffers))
        await super().action_quit()

    def action_show_help_screen(self) -> None:
        self.push_screen(HelpScreen(id="help_screen"))

    def action_toggle_full_screen(self) -> None:
        self.full_screen = not self.full_screen

    def action_toggle_sidebar(self) -> None:
        """
        sidebar_hidden and self.sidebar.disabled both hold important state.
        The sidebar can be hidden with either ctrl+b or f10, and we need
        to persist the state depending on how that happens
        """
        if self.sidebar_hidden is False and self.data_catalog.disabled is True:
            # sidebar was hidden by f10; toggle should show it
            self.data_catalog.disabled = False
        else:
            self.sidebar_hidden = not self.sidebar_hidden

    @work(
        thread=True,
        exclusive=True,
        exit_on_error=False,
        group="query_runners",
        description="Building cursor.",
    )
    def _build_cursors(self, query_text: str) -> None:
        cursors: Dict[str, HarlequinCursor] = {}
        for q in self._split_query_text(query_text):
            cur = self.connection.execute(q)
            if cur is not None:
                if self.run_query_bar.checkbox.value:
                    cur = cur.set_limit(self.limit)
                table_id = f"t{hash(cur)}"
                cursors[table_id] = cur
        self.post_message(QueriesExecuted(query_text=query_text, cursors=cursors))

    def _get_query_text(self) -> str:
        return (
            self._validate_selection()
            or self.editor.current_query
            or self.editor.previous_query
        )

    @staticmethod
    def _split_query_text(query_text: str) -> List[str]:
        return [q for q in query_text.split(";") if q.strip()]

    def _push_error_modal(self, title: str, header: str, error: BaseException) -> None:
        self.push_screen(
            ErrorModal(
                title=title,
                header=header,
                error=error,
            )
        )

    @work(
        thread=True,
        exclusive=True,
        exit_on_error=True,
        group="query_runners",
        description="fetching data from adapter.",
    )
    def _set_result_viewer_data(self, cursors: Dict[str, HarlequinCursor]) -> None:
        errors: List[BaseException] = []
        for id_, cur in cursors.items():
            try:
                cur_data = cur.fetchall()
            except HarlequinQueryError as e:
                errors.append(e)
            else:
                self.call_from_thread(
                    self.results_viewer.push_table,  # type: ignore
                    table_id=id_,
                    column_labels=cur.columns(),
                    data=cur_data,
                )
        elapsed = (
            time.monotonic() - self.query_timer
            if self.query_timer is not None
            else None
        )
        self.post_message(
            ResultsFetched(cursors=cursors, errors=errors, elapsed=elapsed)
        )

    def update_schema_data(self) -> None:
        self._post_updated_catalog()

    @work(thread=True, exclusive=True, exit_on_error=False, group="schema_updaters")
    def _post_updated_catalog(self) -> None:
        catalog = self.connection.get_catalog()
        self.post_message(NewCatalog(catalog=catalog))

    def _validate_selection(self) -> str:
        """
        If the selection is valid query, return it. Otherwise
        return the empty string.
        """
        selection = self.editor.selected_text
        if selection:
            try:
                return self.connection.validate_sql(selection)
            except NotImplementedError:
                return selection
        else:
            return ""



import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union

from platformdirs import user_cache_dir
from textual_textarea.key_handlers import Cursor as Cursor

CACHE_VERSION = 0


@dataclass
class BufferState:
    cursor: Cursor
    selection_anchor: Union[Cursor, None]
    text: str


@dataclass
class Cache:
    focus_index: int  # currently doesn't impact focus on load
    buffers: List[BufferState]


def get_cache_file() -> Path:
    """
    Returns the path to the cache file on disk
    """
    cache_dir = Path(user_cache_dir(appname="harlequin"))
    cache_file = cache_dir / f"cache-{CACHE_VERSION}.pickle"
    return cache_file


def load_cache() -> Union[Cache, None]:
    """
    Returns a Cache (a list of strings) by loading
    from a pickle saved to disk
    """
    cache_file = get_cache_file()
    try:
        with cache_file.open("rb") as f:
            cache: Cache = pickle.load(f)
            assert isinstance(cache, Cache)
    except (
        pickle.UnpicklingError,
        ValueError,
        IndexError,
        FileNotFoundError,
        AssertionError,
    ):
        return None
    else:
        return cache


def write_cache(cache: Cache) -> None:
    """
    Updates dumps buffer contents to to disk
    """
    cache_file = get_cache_file()
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    with open(cache_file, "wb") as f:
        pickle.dump(cache, f)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'harlequin_duckdb': 'src/harlequin_duckdb',
 'harlequin_sqlite': 'src/harlequin_sqlite'}

packages = \
['harlequin', 'harlequin.components', 'harlequin_duckdb', 'harlequin_sqlite']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'duckdb>=0.8.0',
 'platformdirs>=3.10.0,<4.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'questionary>=2.0.1,<3.0.0',
 'rich-click>=1.7.1,<2.0.0',
 'shandy-sqlfmt>=0.19.0',
 'textual-fastdatatable==0.4.0',
 'textual-textarea==0.7.3',
 'textual==0.41.0',
 'tomlkit>=0.12.3,<0.13.0']

extras_require = \
{':python_full_version < "3.10.0"': ['importlib_metadata>=4.6.0'],
 ':python_full_version < "3.11.0"': ['tomli>=2.0.1,<3.0.0'],
 'postgres': ['harlequin-postgres>=0.1,<0.2']}

entry_points = \
{'console_scripts': ['harlequin = harlequin.cli:harlequin'],
 'harlequin.adapter': ['duckdb = harlequin_duckdb:DuckDbAdapter',
                       'sqlite = harlequin_sqlite:HarlequinSqliteAdapter'],
 'pygments.styles': ['harlequin = harlequin.colors:HarlequinPygmentsStyle']}

setup_kwargs = {
    'name': 'harlequin',
    'version': '1.6.0',
    'description': 'The SQL IDE for Your Terminal.',
    'long_description': '# Harlequin\n\n[![PyPI](https://img.shields.io/pypi/v/harlequin)](https://pypi.org/project/harlequin/)\n[![Test](https://github.com/tconbeer/harlequin/actions/workflows/test.yml/badge.svg?branch=main&event=push)](https://github.com/tconbeer/harlequin/actions/workflows/test.yml)\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/harlequin)\n![Runs on Linux | MacOS | Windows](https://img.shields.io/badge/runs%20on-Linux%20%7C%20MacOS%20%7C%20Windows-blue)\n\nThe SQL IDE for Your Terminal.\n\n![Harlequin Demo Video](static/harlequin.gif)\n\n\n## Installing Harlequin\n\nAfter installing Python 3.8 or above, install Harlequin using `pip` or `pipx` with:\n\n```bash\npipx install harlequin\n```\n\n## Using Harlequin\n\nFrom any shell, to open one or more DuckDB database files:\n\n```bash\nharlequin "path/to/duck.db" "another_duck.db"\n```\n\nTo open an in-memory DuckDB session, run Harlequin with no arguments:\n\n```bash\nharlequin\n```\n\n## More info at [harlequin.sh](https://harlequin.sh)\n\nVisit [harlequin.sh](https://harlequin.sh) for an overview of features and full documentation.\n\n## Contributing\n\nThanks for your interest in Harlequin! Harlequin is primarily maintained by [Ted Conbeer](https://github.com/tconbeer), but he welcomes all contributions and is looking for additional maintainers!\n\n### Providing Feedback\n\nWe\'d love to hear from you! [Open an Issue](https://github.com/tconbeer/harlequin/issues/new) to request new features, report bugs, or say hello.\n\n### Setting up Your Dev Environment and Running Tests\n\n1. Install Poetry v1.2 or higher if you don\'t have it already. You may also need or want pyenv, make, and gcc.\n1. Fork this repo, and then clone the fork into a directory (let\'s call it `harlequin`), then `cd harlequin`.\n1. Use `poetry install --sync` to install the project (editable) and its dependencies (including all test and dev dependencies) into a new virtual env.\n1. Use `poetry shell` to spawn a subshell.\n1. Type `make` to run all tests and linters, or run `pytest`, `black .`, `ruff . --fix`, and `mypy` individually.\n\n### Opening PRs\n\n1. PRs should be motivated by an open issue. If there isn\'t already an issue describing the feature or bug, [open one](https://github.com/tconbeer/harlequin/issues/new). Do this before you write code, so you don\'t waste time on something that won\'t get merged.\n2. Ideally new features and bug fixes would be tested, to prevent future regressions. Textual provides a test harness that we use to test features of Harlequin. You can find some examples in the `tests` directory of this project. Please include a test in your PR, but if you can\'t figure it out, open a PR to ask for help.\n2. Open a PR from your fork to the `main` branch of `tconbeer/harlequin`. In the PR description, link to the open issue, and then write a few sentences about **why** you wrote the code you did: explain your design, etc.\n3. Ted may ask you to make changes, or he may make them for you. Don\'t take this the wrong way -- he values your contributions, but he knows this isn\'t your job, either, so if it\'s faster for him, he may push a commit to your branch or create a new branch from your commits.\n',
    'author': 'Ted Conbeer',
    'author_email': 'tconbeer@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://harlequin.sh',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



import re
from typing import List, Union

from rich.text import TextType
from sqlfmt.api import Mode, format_string
from sqlfmt.exception import SqlfmtError
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.message import Message
from textual.widgets import ContentSwitcher, TabbedContent, TabPane, Tabs
from textual_textarea import TextArea, TextAreaSaved
from textual_textarea.key_handlers import Cursor
from textual_textarea.serde import serialize_lines
from textual_textarea.textarea import TextInput

from harlequin.cache import BufferState, load_cache
from harlequin.components.error_modal import ErrorModal


class CodeEditor(TextArea):
    BINDINGS = [
        Binding(
            "ctrl+enter",
            "submit",
            "Run Query",
            key_display="CTRL+ENTER / CTRL+J",
            show=True,
        ),
        Binding("ctrl+j", "submit", "Run Query", show=False),
        Binding("f4", "format", "Format Query", show=True),
    ]

    class Submitted(Message, bubble=True):
        """Posted when user runs the query.

        Attributes:
            lines: The lines of code being submitted.
            cursor: The position of the cursor
        """

        def __init__(self, text: str) -> None:
            super().__init__()
            self.text = text

    @property
    def current_query(self) -> str:
        semicolons = self._semicolons

        if not semicolons:
            return self.text

        before = Cursor(0, 0)
        after: Union[None, Cursor] = None
        for c in semicolons:
            if c <= self.cursor:
                before = c
            elif after is None and c > self.cursor:
                after = c
                break
        else:
            after = Cursor(
                len(self.text_input.lines) - 1, len(self.text_input.lines[-1]) - 1
            )
        return self._get_text_between_cursors(before, after)

    @property
    def previous_query(self) -> str:
        semicolons = self._semicolons

        if not semicolons:
            return self.text

        first = Cursor(0, 0)
        second = Cursor(0, 0)
        for c in semicolons:
            if c <= self.cursor:
                first = second
                second = c
            elif c > self.cursor:
                break

        return self._get_text_between_cursors(first, second)

    def on_mount(self) -> None:
        self.post_message(EditorCollection.EditorSwitched(active_editor=self))
        self.has_shown_clipboard_error = False

    def on_unmount(self) -> None:
        self.post_message(EditorCollection.EditorSwitched(active_editor=None))

    def on_text_area_saved(self, message: TextAreaSaved) -> None:
        self.app.notify(f"Editor contents saved to {message.path}")

    def on_text_area_clipboard_error(self) -> None:
        if not self.has_shown_clipboard_error:
            self.app.notify(
                "Could not access system clipboard.\n"
                "See https://harlequin.sh/docs/troubleshooting#copying-and-pasting",
                severity="error",
                timeout=10,
            )
            self.has_shown_clipboard_error = True

    async def action_submit(self) -> None:
        self.post_message(self.Submitted(self.text))

    def action_format(self) -> None:
        text_input = self.query_one(TextInput)
        old_cursor = text_input.cursor

        try:
            self.text = format_string(self.text, Mode())
        except SqlfmtError as e:
            self.app.push_screen(
                ErrorModal(
                    title="Formatting Error",
                    header="There was an error while formatting your file:",
                    error=e,
                )
            )
        else:
            text_input.move_cursor(old_cursor.pos, old_cursor.lno)
            text_input.update(text_input._content)

    def _get_text_between_cursors(self, before: Cursor, after: Cursor) -> str:
        lines, first, last = self.text_input._get_selected_lines(before, after)
        lines[-1] = lines[-1][: last.pos]
        lines[0] = lines[0][first.pos :]
        return serialize_lines(lines).strip()

    @property
    def _semicolons(self) -> List[Cursor]:
        semicolons: List[Cursor] = []
        for i, line in enumerate(self.text_input.lines):
            for pos in [m.span()[1] for m in re.finditer(";", line)]:
                semicolons.append(Cursor(i, pos))
        return semicolons


class EditorCollection(TabbedContent):
    BINDINGS = [
        Binding("ctrl+n", "new_buffer", "New Tab", show=False),
        Binding("ctrl+w", "close_buffer", "Close Tab", show=False),
        Binding("ctrl+k", "next_buffer", "Next Tab", show=False),
    ]

    BORDER_TITLE = "Query Editor"

    class EditorSwitched(Message):
        def __init__(self, active_editor: Union[CodeEditor, None]) -> None:
            self.active_editor = active_editor
            super().__init__()

    def __init__(
        self,
        *titles: TextType,
        initial: str = "",
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa: A002
        classes: Union[str, None] = None,
        disabled: bool = False,
        language: str = "sql",
        theme: str = "monokai",
    ):
        super().__init__(
            *titles,
            initial=initial,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.language = language
        self.theme = theme
        self.counter = 0

    @property
    def current_editor(self) -> CodeEditor:
        content = self.query_one(ContentSwitcher)
        active_tab_id = self.active
        if active_tab_id:
            try:
                tab_pane = content.query_one(f"#{active_tab_id}", TabPane)
                return tab_pane.query_one(CodeEditor)
            except NoMatches:
                pass
        all_editors = content.query(CodeEditor)
        return all_editors.first(CodeEditor)

    @property
    def all_editors(self) -> List[CodeEditor]:
        content = self.query_one(ContentSwitcher)
        all_editors = content.query(CodeEditor)
        return list(all_editors)

    async def on_mount(self) -> None:
        self.add_class("hide-tabs")
        cache = load_cache()
        if cache is not None:
            for _i, buffer in enumerate(cache.buffers):
                await self.action_new_buffer(state=buffer)
                # we can't load the focus state here, since Tabs
                # really wants to activate the first tab when it's
                # mounted
        else:
            await self.action_new_buffer()
        self.query_one(Tabs).can_focus = False

    def on_focus(self) -> None:
        self.current_editor.focus()

    def on_tabbed_content_tab_activated(
        self, message: TabbedContent.TabActivated
    ) -> None:
        message.stop()
        self.post_message(self.EditorSwitched(active_editor=None))
        self.current_editor.focus()

    async def action_new_buffer(self, state: Union[BufferState, None] = None) -> None:
        self.counter += 1
        new_tab_id = f"tab-{self.counter}"
        editor = CodeEditor(
            id=f"buffer-{self.counter}", language=self.language, theme=self.theme
        )
        pane = TabPane(
            f"Tab {self.counter}",
            editor,
            id=new_tab_id,
        )
        await self.add_pane(pane)  # type: ignore
        if state is not None:
            editor.text = state.text
            editor.cursor = state.cursor
            editor.selection_anchor = state.selection_anchor
        else:
            self.active = new_tab_id
            self.current_editor.focus()
        if self.counter > 1:
            self.remove_class("hide-tabs")

    def action_close_buffer(self) -> None:
        if self.tab_count > 1:
            if self.tab_count == 2:
                self.add_class("hide-tabs")
            self.remove_pane(self.active)
        else:
            self.current_editor.text = ""
            self.current_editor.cursor = (0, 0)  # type: ignore
        self.current_editor.focus()

    def action_next_buffer(self) -> None:
        active = self.active
        if self.tab_count < 2 or active is None:
            return
        tabs = self.query(TabPane)
        next_tabs = tabs[1:]
        next_tabs.append(tabs[0])
        lookup = {t.id: nt.id for t, nt in zip(tabs, next_tabs)}
        self.active = lookup[active]  # type: ignore
        self.post_message(self.EditorSwitched(active_editor=None))
        self.current_editor.focus()
